"""
https://github.github.com/gfm/#code-spans
"""

import pytest

from .utils import act_and_assert


@pytest.mark.gfm
def test_code_spans_338():
    """
    Test case 338:  This is a simple code span:
    """

    # Arrange
    source_markdown = """`foo`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):foo:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_339():
    """
    Test case 339:  Here two backticks are used, because the code contains a backtick. This example also illustrates stripping of a single leading and trailing space:
    """

    # Arrange
    source_markdown = """`` foo ` bar ``"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):foo ` bar:``: : ]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo ` bar</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_340():
    """
    Test case 340:  This example shows the motivation for stripping leading and trailing spaces:
    """

    # Arrange
    source_markdown = """` `` `"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):``:`: : ]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>``</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_341():
    """
    Test case 341:  Note that only one space is stripped:
    """

    # Arrange
    source_markdown = """`  ``  `"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1): `` :`: : ]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code> `` </code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_342():
    """
    Test case 342:  The stripping only happens if the space is on both sides of the string:
    """

    # Arrange
    source_markdown = """` a`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1): a:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code> a</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_343():
    """
    Test case 343:  Only spaces, and not unicode whitespace in general, are stripped in this way:
    """

    # Arrange
    source_markdown = """`\u00A0b\u00A0`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):\u00A0b\u00A0:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>\u00A0b\u00A0</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_344():
    """
    Test case 344:  No stripping occurs if the code span contains only spaces:
    """

    # Arrange
    source_markdown = """` `
`  `"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[icode-span(1,1): :`::]",
        "[text(1,4):\n::\n]",
        "[icode-span(2,1):  :`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code> </code>
<code>  </code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_344a():
    """
    Test case 344a:  Extension of 344 with a whitespace string that is longer than 2.
    """

    # Arrange
    source_markdown = """`   `"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):   :`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>   </code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_345():
    """
    Test case 345:  (part 1) Line endings are treated like spaces:
    """

    # Arrange
    source_markdown = """``
foo
bar\a\a
baz
``""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n\n\n\n]",
        "[icode-span(1,1):foo\a\n\a \abar  \a\n\a \abaz:``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo bar   baz</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_345a():
    """
    Test case 345a:  variations
    """

    # Arrange
    source_markdown = """``
 foo
  bar\a\a
baz
``""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n \n  \n\n]",
        "[icode-span(1,1):foo\a\n\a \abar  \a\n\a \abaz:``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo bar   baz</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_345b():
    """
    Test case 345b:  variations
    """

    # Arrange
    source_markdown = """``this is
a code span``"""
    expected_tokens = [
        "[para(1,1):\n]",
        "[icode-span(1,1):this is\a\n\a \aa code span:``::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>this is a code span</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346():
    """
    Test case 346:  (part 2) Line endings are treated like spaces:
    """

    # Arrange
    source_markdown = """``
foo 
``"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[icode-span(1,1):foo :``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo </code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346ax():
    """
    Test case 346a:  variation
    """

    # Arrange
    source_markdown = """- ``
foo\a
``""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[ulist(1,1):-::2::\n]",
        "[para(1,3):\n\n]",
        "[icode-span(1,3):foo :``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li><code>foo </code></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346aa():
    """
    Test case 346a:  variation
    """

    # Arrange
    source_markdown = """- ``
  foo\a
  ``""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[ulist(1,1):-::2::  \n  ]",
        "[para(1,3):\n\n]",
        "[icode-span(1,3):foo :``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
        "[end-ulist:::True]",
    ]
    expected_gfm = """<ul>
<li><code>foo </code></li>
</ul>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346bx():
    """
    Test case 346a:  variation
    """

    # Arrange
    source_markdown = """> ``
foo 
``"""
    expected_tokens = [
        "[block-quote(1,1)::> \n\n]",
        "[para(1,3):\n\n]",
        "[icode-span(1,3):foo :``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><code>foo </code></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346ba():
    """
    Test case 346a:  variation
    """

    # Arrange
    source_markdown = """> ``
> foo 
> ``"""
    expected_tokens = [
        "[block-quote(1,1)::> \n> \n> ]",
        "[para(1,3):\n\n]",
        "[icode-span(1,3):foo :``:\a\n\a \a:\a\n\a \a]",
        "[end-para:::True]",
        "[end-block-quote:::True]",
    ]
    expected_gfm = """<blockquote>
<p><code>foo </code></p>
</blockquote>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346c():
    """
    Test case 346a:  variation
    """

    # Arrange
    source_markdown = """a[a``
foo\a
``a](/uri)a""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/uri:::::a``\nfoo \n``a:False::::]",
        "[text(1,3):a:]",
        "[icode-span(1,4):foo :``:\a\n\a \a:\a\n\a \a]",
        "[text(3,3):a:]",
        "[end-link::]",
        "[text(3,11):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri">a<code>foo </code>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346ca():
    """
    Test case 346a:  variation
    """

    # Arrange
    source_markdown = """a[a``
 foo\a
 ``a](/uri)a""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n \n ]",
        "[text(1,1):a:]",
        "[link(1,2):inline:/uri:::::a``\nfoo \n``a:False::::]",
        "[text(1,3):a:]",
        "[icode-span(1,4):foo :``:\a\n\a \a:\a\n\a \a]",
        "[text(3,3):a:]",
        "[end-link::]",
        "[text(3,11):a:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>a<a href="/uri">a<code>foo </code>a</a>a</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_346d():
    """
    Test case 346d:  variation
    """

    # Arrange
    source_markdown = """a``
foo 
``a
---"""
    expected_tokens = [
        "[setext(4,1):-:3::(1,1)]",
        "[text(1,1):a:]",
        "[icode-span(1,2):foo :``:\a\n\a \a:\a\n\a \a]",
        "[text(3,3):a:]",
        "[end-setext::]",
    ]
    expected_gfm = """<h2>a<code>foo </code>a</h2>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_347():
    """
    Test case 347:  Interior spaces are not collapsed:
    """

    # Arrange
    source_markdown = """`foo   bar\a
baz`""".replace(
        "\a", " "
    )
    expected_tokens = [
        "[para(1,1):\n]",
        "[icode-span(1,1):foo   bar \a\n\a \abaz:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo   bar  baz</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_348():
    """
    Test case 348:  Note that backslash escapes do not work in code spans. All backslashes are treated literally:
    """

    # Arrange
    source_markdown = """`foo\\`bar`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):foo\\:`::]",
        "[text(1,7):bar`:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo\\</code>bar`</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_349():
    """
    Test case 349:  (part 1) Backslash escapes are never needed, because one can always choose a string of n backtick characters as delimiters, where the code does not contain any strings of exactly n backtick characters.
    """

    # Arrange
    source_markdown = """``foo`bar``"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):foo`bar:``::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo`bar</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_350():
    """
    Test case 350:  (part 2) Backslash escapes are never needed, because one can always choose a string of n backtick characters as delimiters, where the code does not contain any strings of exactly n backtick characters.
    """

    # Arrange
    source_markdown = """` foo `` bar `"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):foo `` bar:`: : ]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>foo `` bar</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_351():
    """
    Test case 351:  Code span backticks have higher precedence than any other inline constructs except HTML tags and autolinks.
    """

    # Arrange
    source_markdown = """*foo`*`"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):*:]",
        "[text(1,2):foo:]",
        "[icode-span(1,5):*:`::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>*foo<code>*</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_352():
    """
    Test case 352:  And this is not parsed as a link:
    """

    # Arrange
    source_markdown = """[not a `link](/foo`)"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):[:]",
        "[text(1,2):not a :]",
        "[icode-span(1,8):link](/foo:`::]",
        "[text(1,20):):]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>[not a <code>link](/foo</code>)</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_353():
    """
    Test case 353:  Code spans, HTML tags, and autolinks have the same precedence. Thus, this is code:
    """

    # Arrange
    source_markdown = """`<a href="`">`"""
    expected_tokens = [
        "[para(1,1):]",
        '[icode-span(1,1):\a<\a&lt;\aa href=\a"\a&quot;\a:`::]',
        '[text(1,12):\a"\a&quot;\a\a>\a&gt;\a`:]',
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>&lt;a href=&quot;</code>&quot;&gt;`</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_354():
    """
    Test case 354:  But this is an HTML tag:
    """

    # Arrange
    source_markdown = """<a href="`">`"""
    expected_tokens = [
        "[para(1,1):]",
        '[raw-html(1,1):a href="`"]',
        "[text(1,13):`:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="`">`</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_355():
    """
    Test case 355:  And this is code:
    """

    # Arrange
    source_markdown = """`<http://foo.bar.`baz>`"""
    expected_tokens = [
        "[para(1,1):]",
        "[icode-span(1,1):\a<\a&lt;\ahttp://foo.bar.:`::]",
        "[text(1,19):baz\a>\a&gt;\a`:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><code>&lt;http://foo.bar.</code>baz&gt;`</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_356():
    """
    Test case 356:  But this is an autolink:
    """

    # Arrange
    source_markdown = """<http://foo.bar.`baz>`"""
    expected_tokens = [
        "[para(1,1):]",
        "[uri-autolink(1,1):http://foo.bar.`baz]",
        "[text(1,22):`:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p><a href="http://foo.bar.%60baz">http://foo.bar.`baz</a>`</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_357():
    """
    Test case 357:  (part 1) When a backtick string is not closed by a matching backtick string, we just have literal backticks:
    """

    # Arrange
    source_markdown = """```foo``"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):```foo``:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>```foo``</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_358():
    """
    Test case 358:  (part 2) When a backtick string is not closed by a matching backtick string, we just have literal backticks:
    """

    # Arrange
    source_markdown = """`foo"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):`foo:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>`foo</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_359():
    """
    Test case 359:  The following case also illustrates the need for opening and closing backtick strings to be equal in length:
    """

    # Arrange
    source_markdown = """`foo``bar``"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):`foo:]",
        "[icode-span(1,5):bar:``::]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>`foo<code>bar</code></p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_extra_01():
    """
    Test case 115:
    """

    # Arrange
    source_markdown = """aa``` a
a ```aa
foo"""
    expected_tokens = [
        "[para(1,1):\n\n]",
        "[text(1,1):aa:]",
        "[icode-span(1,3):a\a\n\a \aa:```: : ]",
        "[text(2,6):aa\nfoo::\n]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>aa<code>a a</code>aa
foo</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_extra_02():
    """
    Test case 02:  Paragraph with single code span start and no space outside and inside
    """

    # Arrange
    source_markdown = """aa`aa`aa"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):aa:]",
        "[icode-span(1,3):aa:`::]",
        "[text(1,7):aa:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>aa<code>aa</code>aa</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_extra_03():
    """
    Test case 03:  Paragraph with double code span start and one space outside and inside
    """

    # Arrange
    source_markdown = """aa `` aa `` aa"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):aa :]",
        "[icode-span(1,4):aa:``: : ]",
        "[text(1,12): aa:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>aa <code>aa</code> aa</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)


@pytest.mark.gfm
def test_code_spans_extra_04():
    """
    Test case 04:  Paragraph with triple code span start and three spaces outside and inside
    """

    # Arrange
    source_markdown = """aa  ```  aa  ```  aa"""
    expected_tokens = [
        "[para(1,1):]",
        "[text(1,1):aa  :]",
        "[icode-span(1,5): aa :```: : ]",
        "[text(1,17):  aa:]",
        "[end-para:::True]",
    ]
    expected_gfm = """<p>aa  <code> aa </code>  aa</p>"""

    # Act & Assert
    act_and_assert(source_markdown, expected_gfm, expected_tokens)



"""
Module to provide helper functions for parsing html.
"""
import logging
import string

from pymarkdown.constants import Constants
from pymarkdown.inline_markdown_token import RawHtmlMarkdownToken, TextMarkdownToken
from pymarkdown.leaf_markdown_token import HtmlBlockMarkdownToken
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.parser_logger import ParserLogger
from pymarkdown.stack_token import HtmlBlockStackToken, ParagraphStackToken

POGGER = ParserLogger(logging.getLogger(__name__))


class HtmlHelper:
    """
    Class to provide helper functions for parsing html.
    """

    html_block_1 = "1"
    html_block_2 = "2"
    html_block_3 = "3"
    html_block_4 = "4"
    html_block_5 = "5"
    html_block_6 = "6"
    html_block_7 = "7"

    __html_block_start_character = "<"
    __html_tag_start = "/"
    __html_tag_name_end = " >"
    __html_tag_end = ">"
    __html_attribute_value_single = "'"
    __html_attribute_value_double = '"'
    __html_attribute_name_value_separator = "="
    __html_attribute_separator = ParserHelper.space_character
    __valid_tag_name_start = string.ascii_letters
    __valid_tag_name_characters = f"{string.ascii_letters}{string.digits}-"
    __tag_attribute_name_characters = f"{string.ascii_letters}{string.digits}_.:-"
    __unquoted_attribute_value_stop = f"\"'=<>`{Constants.whitespace}"
    __tag_attribute_name_start = f"{string.ascii_letters}_:"
    __html_block_1_start_tag_prefix = ["script", "pre", "style"]
    __html_tag_attribute_value_terminators = " \"'=<>`"
    __html_block_2_to_5_start = "!"
    __html_block_2_continued_start = "--"
    __html_block_2_xx = f"{__html_block_2_to_5_start}{__html_block_2_continued_start}"
    __html_block_3_continued_start = "?"
    __html_block_4_continued_start = string.ascii_uppercase
    __html_block_5_continued_start = "[CDATA["
    __html_block_5_xx = f"{__html_block_2_to_5_start}{__html_block_5_continued_start}"
    __html_block_1_end_tags = ["</script>", "</pre>", "</style>"]
    __html_block_2_end = "-->"
    __html_block_3_end = "?>"
    __html_block_4_end = __html_tag_end
    __html_block_5_end = "]]>"

    __attribute_start_characters = "abcdefghijklmnopqrstuvwxyz1234567890:_"
    __attribute_other_characters = f"{__attribute_start_characters}.-"

    __raw_declaration_start_character = "!"
    __raw_declaration_whitespace = ParserHelper.space_character
    __raw_html_exclusion_1 = ">"
    __raw_html_exclusion_2 = "->"
    __raw_html_exclusion_3 = "-"
    __raw_html_exclusion_4 = "--"

    __html_block_6_start = [
        "address",
        "article",
        "aside",
        "base",
        "basefont",
        "blockquote",
        "body",
        "caption",
        "center",
        "col",
        "colgroup",
        "dd",
        "details",
        "dialog",
        "dir",
        "div",
        "dl",
        "dt",
        "fieldset",
        "figcaption",
        "figure",
        "footer",
        "form",
        "frame",
        "frameset",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "head",
        "header",
        "hr",
        "html",
        "iframe",
        "legend",
        "li",
        "link",
        "main",
        "menu",
        "menuitem",
        "nav",
        "noframes",
        "ol",
        "optgroup",
        "option",
        "p",
        "param",
        "section",
        "source",
        "summary",
        "table",
        "tbody",
        "td",
        "tfoot",
        "th",
        "thead",
        "title",
        "tr",
        "track",
        "ul",
    ]

    @staticmethod
    def is_valid_tag_name(tag_name):
        """
        Determine if the html tag name is valid according to the html rules.
        """

        if not tag_name:
            return False
        for next_character in tag_name.lower():
            if next_character not in HtmlHelper.__valid_tag_name_characters:
                return False
        return True

    @staticmethod
    def extract_html_attribute_name(string_to_parse, string_index):
        """
        Attempt to extract the attribute name from the provided string.
        """

        string_to_parse_length = len(string_to_parse)
        if not (
            string_index < string_to_parse_length
            and (
                string_to_parse[string_index] in HtmlHelper.__attribute_start_characters
            )
        ):
            return -1
        string_index += 1
        while string_index < string_to_parse_length:
            if not (
                string_to_parse[string_index] in HtmlHelper.__attribute_other_characters
            ):
                break
            string_index += 1

        if string_index < string_to_parse_length and (
            string_to_parse[string_index]
            == HtmlHelper.__html_attribute_name_value_separator
            or string_to_parse[string_index] == HtmlHelper.__html_attribute_separator
            or string_to_parse[string_index] == HtmlHelper.__html_tag_start
            or string_to_parse[string_index] == HtmlHelper.__html_tag_end
        ):
            return string_index
        return -1

    @staticmethod
    def extract_optional_attribute_value(line_to_parse, value_index):
        """
        Determine and extract an optional attribute value.
        """

        non_whitespace_index, _ = ParserHelper.extract_whitespace(
            line_to_parse, value_index
        )
        line_to_parse_size = len(line_to_parse)
        if (
            non_whitespace_index < line_to_parse_size
            and line_to_parse[non_whitespace_index]
            != HtmlHelper.__html_attribute_name_value_separator
        ) or non_whitespace_index >= line_to_parse_size:
            return non_whitespace_index

        non_whitespace_index, _ = ParserHelper.extract_whitespace(
            line_to_parse, non_whitespace_index + 1
        )
        if non_whitespace_index < line_to_parse_size:
            first_character_of_value = line_to_parse[non_whitespace_index]
            if first_character_of_value == HtmlHelper.__html_attribute_value_double:
                (
                    non_whitespace_index,
                    extracted_text,
                ) = ParserHelper.collect_until_character(
                    line_to_parse,
                    non_whitespace_index + 1,
                    HtmlHelper.__html_attribute_value_double,
                )
                if non_whitespace_index == line_to_parse_size:
                    return -1
                non_whitespace_index += 1
            elif first_character_of_value == HtmlHelper.__html_attribute_value_single:
                (
                    non_whitespace_index,
                    extracted_text,
                ) = ParserHelper.collect_until_character(
                    line_to_parse,
                    non_whitespace_index + 1,
                    HtmlHelper.__html_attribute_value_single,
                )
                if non_whitespace_index == line_to_parse_size:
                    return -1
                non_whitespace_index += 1
            else:
                (
                    non_whitespace_index,
                    extracted_text,
                ) = ParserHelper.collect_until_one_of_characters(
                    line_to_parse,
                    non_whitespace_index,
                    HtmlHelper.__html_tag_attribute_value_terminators,
                )

                if not extracted_text:
                    non_whitespace_index = -1
        else:
            non_whitespace_index = -1
        return non_whitespace_index

    @staticmethod
    def is_complete_html_end_tag(tag_name, line_to_parse, next_char_index):
        """
        Determine if the supplied information is a completed end of tag specification.
        """

        is_valid = HtmlHelper.is_valid_tag_name(tag_name)
        non_whitespace_index, _ = ParserHelper.extract_whitespace(
            line_to_parse, next_char_index
        )
        have_end_of_tag = (
            non_whitespace_index < len(line_to_parse)
            and line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_end
        )
        return have_end_of_tag and is_valid, non_whitespace_index + 1

    @staticmethod
    def __is_valid_block_1_tag_name(tag_name):
        """
        Determine if the tag name is a valid block-1 html tag name.
        """

        return tag_name in HtmlHelper.__html_block_1_start_tag_prefix

    @staticmethod
    def is_complete_html_start_tag(tag_name, line_to_parse, next_char_index):
        """
        Determine if the supplied information is a completed start of tag specification.
        """

        is_tag_valid = HtmlHelper.is_valid_tag_name(
            tag_name
        ) and not HtmlHelper.__is_valid_block_1_tag_name(tag_name)

        non_whitespace_index, extracted_whitespace = ParserHelper.extract_whitespace(
            line_to_parse, next_char_index
        )

        are_attributes_valid, line_to_parse_size = True, len(line_to_parse)
        while (
            is_tag_valid
            and extracted_whitespace
            and are_attributes_valid
            and (0 <= non_whitespace_index < line_to_parse_size)
            and not (
                line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_end
                or line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_start
            )
        ):

            non_whitespace_index = HtmlHelper.extract_html_attribute_name(
                line_to_parse, non_whitespace_index
            )
            if non_whitespace_index == -1:
                are_attributes_valid = False
                break
            non_whitespace_index = HtmlHelper.extract_optional_attribute_value(
                line_to_parse, non_whitespace_index
            )
            if non_whitespace_index == -1:
                are_attributes_valid = False
                break
            (
                non_whitespace_index,
                extracted_whitespace,
            ) = ParserHelper.extract_whitespace(line_to_parse, non_whitespace_index)

        is_end_of_tag_present = False
        if non_whitespace_index < line_to_parse_size:
            if line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_start:
                non_whitespace_index += 1
            if line_to_parse[non_whitespace_index] == HtmlHelper.__html_tag_end:
                non_whitespace_index += 1
                is_end_of_tag_present = True

        non_whitespace_index, _ = ParserHelper.extract_whitespace(
            line_to_parse, non_whitespace_index
        )
        return (
            (
                is_tag_valid
                and is_end_of_tag_present
                and non_whitespace_index == line_to_parse_size
                and are_attributes_valid
            ),
            non_whitespace_index,
        )

    @staticmethod
    def __parse_raw_tag_name(text_to_parse, start_index):
        """
        Parse a HTML tag name from the string.
        """
        if ParserHelper.is_character_at_index_one_of(
            text_to_parse, start_index, HtmlHelper.__valid_tag_name_start
        ):
            index = start_index + 1
            while ParserHelper.is_character_at_index_one_of(
                text_to_parse, index, HtmlHelper.__valid_tag_name_characters
            ):
                index += 1
            tag_name = text_to_parse[0:index]
        else:
            tag_name = ""
        return tag_name

    @staticmethod
    def __parse_tag_attributes(text_to_parse, start_index):
        """
        Handle the parsing of the attributes for an open tag.
        """
        parse_index, _ = ParserHelper.collect_while_one_of_characters(
            text_to_parse, start_index, HtmlHelper.__tag_attribute_name_characters
        )
        end_name_index, extracted_whitespace = ParserHelper.extract_any_whitespace(
            text_to_parse, parse_index
        )
        if ParserHelper.is_character_at_index(
            text_to_parse,
            end_name_index,
            HtmlHelper.__html_attribute_name_value_separator,
        ):
            (
                value_start_index,
                extracted_whitespace,
            ) = ParserHelper.extract_any_whitespace(text_to_parse, end_name_index + 1)
            if ParserHelper.is_character_at_index_one_of(
                text_to_parse,
                value_start_index,
                HtmlHelper.__html_attribute_value_single,
            ):
                value_end_index, _ = ParserHelper.collect_until_character(
                    text_to_parse,
                    value_start_index + 1,
                    HtmlHelper.__html_attribute_value_single,
                )
                if not ParserHelper.is_character_at_index(
                    text_to_parse,
                    value_end_index,
                    HtmlHelper.__html_attribute_value_single,
                ):
                    return None, -1
                value_end_index += 1
            elif ParserHelper.is_character_at_index_one_of(
                text_to_parse,
                value_start_index,
                HtmlHelper.__html_attribute_value_double,
            ):
                value_end_index, _ = ParserHelper.collect_until_character(
                    text_to_parse,
                    value_start_index + 1,
                    HtmlHelper.__html_attribute_value_double,
                )
                if not ParserHelper.is_character_at_index(
                    text_to_parse,
                    value_end_index,
                    HtmlHelper.__html_attribute_value_double,
                ):
                    return None, -1
                value_end_index += 1
            else:
                value_end_index, _ = ParserHelper.collect_until_one_of_characters(
                    text_to_parse,
                    value_start_index,
                    HtmlHelper.__unquoted_attribute_value_stop,
                )
            end_name_index, extracted_whitespace = ParserHelper.extract_any_whitespace(
                text_to_parse, value_end_index
            )

        return end_name_index, extracted_whitespace

    @staticmethod
    def __parse_raw_open_tag(text_to_parse):
        """
        Parse the current line as if it is an open tag, and determine if it is valid.
        """

        end_parse_index, valid_raw_html, tag_name = (
            -1,
            None,
            HtmlHelper.__parse_raw_tag_name(text_to_parse, 0),
        )
        if tag_name:
            parse_index, extracted_whitespace = ParserHelper.extract_any_whitespace(
                text_to_parse, len(tag_name)
            )
            if extracted_whitespace:
                while (
                    extracted_whitespace
                    and ParserHelper.is_character_at_index_one_of(
                        text_to_parse,
                        parse_index,
                        HtmlHelper.__tag_attribute_name_start,
                    )
                ):
                    (
                        parse_index,
                        extracted_whitespace,
                    ) = HtmlHelper.__parse_tag_attributes(text_to_parse, parse_index)
                    if parse_index is None:
                        return parse_index, extracted_whitespace

            if ParserHelper.is_character_at_index(
                text_to_parse, parse_index, HtmlHelper.__html_tag_start
            ):
                parse_index += 1

            if ParserHelper.is_character_at_index(
                text_to_parse, parse_index, HtmlHelper.__html_tag_end
            ):
                valid_raw_html = text_to_parse[0:parse_index]
                end_parse_index = parse_index + 1

        return valid_raw_html, end_parse_index

    @staticmethod
    def __parse_raw_close_tag(text_to_parse):
        """
        Parse the current line as if it is a close tag, and determine if it is valid.
        """
        valid_raw_html = None
        if ParserHelper.is_character_at_index(
            text_to_parse, 0, HtmlHelper.__html_tag_start
        ):
            tag_name = HtmlHelper.__parse_raw_tag_name(text_to_parse, 1)
            if tag_name:
                parse_index, text_to_parse_size = len(tag_name), len(text_to_parse)
                if parse_index != text_to_parse_size:
                    parse_index, _ = ParserHelper.extract_whitespace(
                        text_to_parse, parse_index
                    )
                if parse_index == text_to_parse_size:
                    valid_raw_html = text_to_parse
        return valid_raw_html

    @staticmethod
    def __parse_raw_declaration(text_to_parse):
        """
        Parse a possible raw html declaration sequence, and return if it is valid.
        """

        valid_raw_html = None
        if ParserHelper.is_character_at_index_one_of(
            text_to_parse, 0, HtmlHelper.__raw_declaration_start_character
        ):
            (
                parse_index,
                declaration_name,
            ) = ParserHelper.collect_while_one_of_characters(
                text_to_parse, 1, HtmlHelper.__html_block_4_continued_start
            )
            if declaration_name:
                whitespace_count, _ = ParserHelper.collect_while_character(
                    text_to_parse, parse_index, HtmlHelper.__raw_declaration_whitespace
                )
                if whitespace_count:
                    valid_raw_html = text_to_parse
        return valid_raw_html

    @staticmethod
    def __process_raw_special(
        remaining_line,
        special_start,
        special_end,
        do_extra_check=False,
    ):
        """
        Parse a possible raw html special sequence, and return if it is valid.
        """
        valid_raw_html, parse_index, special_start_size = None, -1, len(special_start)
        if remaining_line.startswith(special_start):
            remaining_line = remaining_line[special_start_size:]
            parse_index = remaining_line.find(special_end)
            if parse_index != -1:
                remaining_line = remaining_line[0:parse_index]
                parse_index = parse_index + special_start_size + len(special_end)
                if (not do_extra_check) or (
                    not (
                        remaining_line[0] == HtmlHelper.__raw_html_exclusion_1
                        or remaining_line.startswith(HtmlHelper.__raw_html_exclusion_2)
                        or remaining_line[-1] == HtmlHelper.__raw_html_exclusion_3
                        or HtmlHelper.__raw_html_exclusion_4 in remaining_line
                    )
                ):
                    valid_raw_html = (
                        f"{special_start}{remaining_line}{special_end[0:-1]}"
                    )
        return valid_raw_html, parse_index

    @staticmethod
    def parse_raw_html(
        only_between_angles, remaining_line, line_number, column_number, inline_request
    ):
        """
        Given an open HTML tag character (<), try the various possibilities for
        types of tag, and determine if any of them parse validly.
        """

        valid_raw_html, remaining_line_parse_index = HtmlHelper.__parse_raw_open_tag(
            remaining_line
        )
        if not valid_raw_html:
            valid_raw_html = HtmlHelper.__parse_raw_close_tag(only_between_angles)
        if not valid_raw_html:
            (
                valid_raw_html,
                remaining_line_parse_index,
            ) = HtmlHelper.__process_raw_special(
                remaining_line,
                HtmlHelper.__html_block_2_xx,
                HtmlHelper.__html_block_2_end,
                True,
            )
        if not valid_raw_html:
            (
                valid_raw_html,
                remaining_line_parse_index,
            ) = HtmlHelper.__process_raw_special(
                remaining_line,
                HtmlHelper.__html_block_3_continued_start,
                HtmlHelper.__html_block_3_end,
            )
        if not valid_raw_html:
            (
                valid_raw_html,
                remaining_line_parse_index,
            ) = HtmlHelper.__process_raw_special(
                remaining_line,
                HtmlHelper.__html_block_5_xx,
                HtmlHelper.__html_block_5_end,
            )
        if not valid_raw_html:
            valid_raw_html = HtmlHelper.__parse_raw_declaration(only_between_angles)

        if valid_raw_html:
            if inline_request.para_owner:
                (
                    valid_raw_html,
                    inline_request.para_owner.rehydrate_index,
                ) = ParserHelper.recombine_string_with_whitespace(
                    valid_raw_html,
                    inline_request.para_owner.extracted_whitespace,
                    inline_request.para_owner.rehydrate_index,
                    add_replace_marker_if_empty=True,
                )
            return (
                RawHtmlMarkdownToken(valid_raw_html, line_number, column_number),
                remaining_line_parse_index,
            )
        return None, -1

    @staticmethod
    def __check_for_special_html_blocks(line_to_parse, character_index):
        """
        Check for the easy to spot special blocks: 2-5.
        """

        html_block_type = None
        if character_index < len(line_to_parse):
            if ParserHelper.is_character_at_index(
                line_to_parse, character_index, HtmlHelper.__html_block_2_to_5_start
            ):
                if ParserHelper.are_characters_at_index(
                    line_to_parse,
                    character_index + 1,
                    HtmlHelper.__html_block_2_continued_start,
                ):
                    html_block_type = HtmlHelper.html_block_2
                elif ParserHelper.is_character_at_index_one_of(
                    line_to_parse,
                    character_index + 1,
                    HtmlHelper.__html_block_4_continued_start,
                ):
                    html_block_type = HtmlHelper.html_block_4
                elif ParserHelper.are_characters_at_index(
                    line_to_parse,
                    character_index + 1,
                    HtmlHelper.__html_block_5_continued_start,
                ):
                    html_block_type = HtmlHelper.html_block_5
            elif ParserHelper.is_character_at_index(
                line_to_parse,
                character_index,
                HtmlHelper.__html_block_3_continued_start,
            ):
                html_block_type = HtmlHelper.html_block_3

        return html_block_type

    @staticmethod
    def __check_for_normal_html_blocks(
        remaining_html_tag, line_to_parse, character_index
    ):
        """
        Check for the the html blocks that are harder to identify: 1, 6-7.
        """

        html_block_type = None

        if HtmlHelper.__is_valid_block_1_tag_name(remaining_html_tag):
            html_block_type = HtmlHelper.html_block_1
        else:
            adjusted_remaining_html_tag, line_to_parse_size = remaining_html_tag, len(
                line_to_parse
            )
            is_end_tag = (
                adjusted_remaining_html_tag
                and adjusted_remaining_html_tag[0] == HtmlHelper.__html_tag_start
            )
            if is_end_tag:
                adjusted_remaining_html_tag = adjusted_remaining_html_tag[1:]

            if (
                character_index < line_to_parse_size
                and line_to_parse[character_index] == HtmlHelper.__html_tag_end
                and adjusted_remaining_html_tag
                and adjusted_remaining_html_tag[-1] == HtmlHelper.__html_tag_start
            ):
                adjusted_remaining_html_tag = adjusted_remaining_html_tag[0:-1]
            if adjusted_remaining_html_tag in HtmlHelper.__html_block_6_start:
                html_block_type = HtmlHelper.html_block_6
            elif is_end_tag:
                is_complete, complete_parse_index = HtmlHelper.is_complete_html_end_tag(
                    adjusted_remaining_html_tag, line_to_parse, character_index
                )
                if is_complete:
                    html_block_type = HtmlHelper.html_block_7
                    character_index = complete_parse_index
            else:
                (
                    is_complete,
                    complete_parse_index,
                ) = HtmlHelper.is_complete_html_start_tag(
                    adjusted_remaining_html_tag, line_to_parse, character_index
                )
                if is_complete:
                    html_block_type = HtmlHelper.html_block_7
                    character_index = complete_parse_index
            if html_block_type == HtmlHelper.html_block_7:
                new_index, _ = ParserHelper.extract_whitespace(
                    line_to_parse, character_index
                )
                if new_index != line_to_parse_size:
                    html_block_type = None
        return html_block_type

    @staticmethod
    def __determine_html_block_type(token_stack, line_to_parse, start_index):
        """
        Determine the type of the html block that we are starting.
        """

        character_index, remaining_html_tag = start_index + 1, ""
        html_block_type = HtmlHelper.__check_for_special_html_blocks(
            line_to_parse, character_index
        )
        if not html_block_type:
            (
                character_index,
                remaining_html_tag,
            ) = ParserHelper.collect_until_one_of_characters(
                line_to_parse, character_index, HtmlHelper.__html_tag_name_end
            )
            remaining_html_tag = remaining_html_tag.lower()

            html_block_type = HtmlHelper.__check_for_normal_html_blocks(
                remaining_html_tag, line_to_parse, character_index
            )

        POGGER.debug("html_block_type=$", html_block_type)
        if not html_block_type:
            return None, None
        if html_block_type == HtmlHelper.html_block_7:
            if token_stack[-1].is_paragraph:
                POGGER.debug("html_block_type 7 cannot interrupt a paragraph")
                return None, None
        return html_block_type, remaining_html_tag

    @staticmethod
    def is_html_block(line_to_parse, start_index, extracted_whitespace, token_stack):
        """
        Determine if the current sequence of characters would start a html block element.
        """

        html_block_type, remaining_html_tag = None, None
        if (
            ParserHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
        ) and ParserHelper.is_character_at_index(
            line_to_parse,
            start_index,
            HtmlHelper.__html_block_start_character,
        ):
            (
                html_block_type,
                remaining_html_tag,
            ) = HtmlHelper.__determine_html_block_type(
                token_stack,
                line_to_parse,
                start_index,
            )
        return html_block_type, remaining_html_tag

    @staticmethod
    def parse_html_block(parser_state, position_marker, extracted_whitespace):
        """
        Determine if we have the criteria that we need to start an HTML block.
        """

        new_tokens = []
        html_block_type, _ = HtmlHelper.is_html_block(
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
            parser_state.token_stack,
        )
        if html_block_type:
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[ParagraphStackToken],
            )
            new_token = HtmlBlockMarkdownToken(position_marker, extracted_whitespace)
            new_tokens.append(new_token)
            parser_state.token_stack.append(
                HtmlBlockStackToken(html_block_type, new_token)
            )
        return new_tokens

    @staticmethod
    def check_blank_html_block_end(parser_state):
        """
        Check to see if we have encountered the end of the current HTML block
        via an empty line or BLANK.
        """

        new_tokens = []
        if (
            parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_6
            or parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_7
        ):
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[type(parser_state.token_stack[-1])],
            )

        return new_tokens

    @staticmethod
    def check_normal_html_block_end(
        parser_state, line_to_parse, start_index, extracted_whitespace, position_marker
    ):
        """
        Check to see if we have encountered the end of the current HTML block
        via text on a normal line.
        """

        new_tokens = [
            TextMarkdownToken(
                line_to_parse[start_index:],
                extracted_whitespace,
                position_marker=position_marker,
            )
        ]

        is_block_terminated, adj_line = False, line_to_parse[start_index:]
        if parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_1:
            for next_end_tag in HtmlHelper.__html_block_1_end_tags:
                if next_end_tag in adj_line:
                    is_block_terminated = True
        elif parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_2:
            is_block_terminated = HtmlHelper.__html_block_2_end in adj_line
        elif parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_3:
            is_block_terminated = HtmlHelper.__html_block_3_end in adj_line
        elif parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_4:
            is_block_terminated = HtmlHelper.__html_block_4_end in adj_line
        elif parser_state.token_stack[-1].html_block_type == HtmlHelper.html_block_5:
            is_block_terminated = HtmlHelper.__html_block_5_end in adj_line

        if is_block_terminated:
            terminated_block_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[type(parser_state.token_stack[-1])],
            )
            assert terminated_block_tokens
            new_tokens.extend(terminated_block_tokens)
        return new_tokens



"""
Module to implement a plugin that looks for multiple spaces after the hash
mark on an atx heading.
"""
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd019(Plugin):
    """
    Class to implement a plugin that looks for multiple spaces after the hash
    mark on an atx heading.
    """

    def __init__(self):
        super().__init__()
        self.__atx_heading_token = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-multiple-space-atx",
            plugin_id="MD019",
            plugin_enabled_by_default=True,
            plugin_description="Multiple spaces are present after hash character on Atx Heading.",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md019.md",
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__atx_heading_token = None

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_atx_heading:
            if not token.remove_trailing_count:
                self.__atx_heading_token = token
        elif token.is_paragraph_end:
            self.__atx_heading_token = None
        elif token.is_text:
            resolved_extracted_whitespace = ParserHelper.remove_all_from_text(
                token.extracted_whitespace
            )
            if self.__atx_heading_token and len(resolved_extracted_whitespace) > 1:
                self.report_next_token_error(context, self.__atx_heading_token)



"""
Module to provide tests related to the MD019 rule.
"""
from test.markdown_scanner import MarkdownScanner

import pytest

# pylint: disable=too-many-lines


@pytest.mark.rules
def test_md019_good_single_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md019 directory that has atx headings with a single space
    after the initial hash.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md019/single_spacing.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


@pytest.mark.rules
def test_md019_bad_multiple_spacing():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md019 directory that has atx headings with multiple spaces
    after the initial hash.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md019/multiple_spacing.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md019/multiple_spacing.md:1:1: "
        + "MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)\n"
        "test/resources/rules/md019/multiple_spacing.md:3:1: "
        + "MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_md019_bad_multiple_spacing_with_inline():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md019 directory that has atx headings with a single space
    after the initial hash, with inline processing in the heading for good measure.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "scan",
        "test/resources/rules/md019/multiple_spacing_with_inline.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md019/multiple_spacing_with_inline.md:1:1: "
        + "MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)\n"
        "test/resources/rules/md019/multiple_spacing_with_inline.md:3:1: "
        + "MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_md019_bad_multiple_spacing_with_indent():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md019 directory that has atx headings with a single space
    after the initial hash, with inline processing in the heading for good measure.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md023",
        "scan",
        "test/resources/rules/md019/multiple_spacing_with_indent.md",
    ]

    expected_return_code = 1
    expected_output = (
        "test/resources/rules/md019/multiple_spacing_with_indent.md:1:2: "
        + "MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)\n"
        "test/resources/rules/md019/multiple_spacing_with_indent.md:3:3: "
        + "MD019: Multiple spaces are present after hash character on Atx Heading. (no-multiple-space-atx)\n"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_md019_bad_single_space_single_tab():
    """
    Test to make sure we get the expected behavior after scanning a good file from the
    test/resources/rules/md019 directory that has two headers that start with various
    amounts of leading whitespace, followed by 1-2 hash characters, a single space
    character, and a single tab character.  This rule should not fire as in both cases,
    there is a single space after the hash character.
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = [
        "--disable-rules",
        "md010",
        "scan",
        "test/resources/rules/md019/single_space_single_tab.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Module to implement a plugin that looks for excessive spaces after the block quote character.
"""
from pymarkdown.parser_helper import ParserHelper
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd027(Plugin):
    """
    Class to implement a plugin that looks for excessive spaces after the block quote character.
    """

    def __init__(self):
        super().__init__()
        self.__container_tokens = None
        self.__bq_line_index = None
        self.__last_leaf_token = None
        self.__line_index_at_bq_start = None
        self.__debug_on = False

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-multiple-space-blockquote",
            plugin_id="MD027",
            plugin_enabled_by_default=True,
            plugin_description="Multiple spaces after blockquote symbol",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md027.md",
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__container_tokens = []
        self.__bq_line_index = {}
        self.__last_leaf_token = None
        self.__line_index_at_bq_start = None

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        # if self.__debug_on:
        #     print("token>" + ParserHelper.make_value_visible(token))
        if token.is_block_quote_start:
            self.__container_tokens.append(token)
            self.__bq_line_index[len(self.__container_tokens)] = 0
            # if self.__debug_on:
            #     print("bq>" + str(self.__container_tokens[-1]).replace("\n","\\n"))
        elif token.is_block_quote_end:
            num_container_tokens = len(self.__container_tokens)
            newlines_in_container = self.__container_tokens[-1].leading_spaces.count(
                "\n"
            )
            if self.__container_tokens[
                -1
            ].leading_spaces and not self.__container_tokens[
                -1
            ].leading_spaces.endswith(
                "\n"
            ):
                newlines_in_container += 1

            # assert (
            #     newlines_in_container == self.__bq_line_index[num_container_tokens]
            # ), (
            #     str(newlines_in_container)
            #     + " == "
            #     + str(self.__bq_line_index[num_container_tokens])
            # )
            del self.__bq_line_index[num_container_tokens]
            del self.__container_tokens[-1]
        elif token.is_list_start:
            self.__container_tokens.append(token)
        elif token.is_list_end:
            del self.__container_tokens[-1]
        elif (
            self.__container_tokens and self.__container_tokens[-1].is_block_quote_start
        ):
            self.__handle_within_block_quotes(context, token)

    def __handle_blank_line(self, context, token, num_container_tokens):
        if token.extracted_whitespace:
            # if self.__debug_on:
            #     print("blank-error")
            self.report_next_token_error(context, token)
        self.__bq_line_index[num_container_tokens] += 1

    def __handle_thematic_break(self, context, token, num_container_tokens):
        if token.extracted_whitespace:
            column_number_delta = -(
                token.column_number - len(token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("thematic-end-error")
            self.report_next_token_error(
                context, token, column_number_delta=column_number_delta
            )
        self.__bq_line_index[num_container_tokens] += 1

    def __handle_atx_heading(self, context, token, num_container_tokens):
        if token.extracted_whitespace:
            column_number_delta = -(
                token.column_number - len(token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("atx-error")
            self.report_next_token_error(
                context, token, column_number_delta=column_number_delta
            )
        self.__last_leaf_token = token
        self.__bq_line_index[num_container_tokens] += 1

    def __handle_setext_heading(self, context, token):
        if token.extracted_whitespace:
            line_number_delta = token.original_line_number - token.line_number
            column_number_delta = -(
                token.original_column_number - len(token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("setext-error")
            #     print("line->" + str(token.original_line_number) + ",col=" + str(token.original_column_number))
            #     print("delta->line->" + str(line_number_delta) + ",col=" + str(column_number_delta))
            self.report_next_token_error(
                context,
                token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
            )
        self.__last_leaf_token = token

    def __handle_setext_heading_end(self, context, token, num_container_tokens):
        if token.extracted_whitespace:
            column_number_delta = -(
                self.__last_leaf_token.column_number - len(token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("setext-end-error")
            self.report_next_token_error(
                context, self.__last_leaf_token, column_number_delta=column_number_delta
            )
        self.__last_leaf_token = None
        self.__bq_line_index[num_container_tokens] += 1

    def __handle_fenced_code_block(self, context, token, num_container_tokens):
        if token.extracted_whitespace:
            column_number_delta = -(
                token.column_number - len(token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("fenced-start-error")
            self.report_next_token_error(
                context, token, column_number_delta=column_number_delta
            )
        self.__last_leaf_token = token
        self.__bq_line_index[num_container_tokens] += 1
        self.__line_index_at_bq_start = self.__bq_line_index[num_container_tokens]

    def __handle_fenced_code_block_end(self, context, token, num_container_tokens):
        if token.extracted_whitespace:
            scoped_block_quote_token = self.__container_tokens[-1]
            split_leading_spaces = scoped_block_quote_token.leading_spaces.split("\n")
            specific_block_quote_prefix = split_leading_spaces[
                self.__bq_line_index[num_container_tokens]
            ]

            line_number_delta = 1 + (
                self.__bq_line_index[num_container_tokens]
                - self.__line_index_at_bq_start
            )
            column_number_delta = -(len(specific_block_quote_prefix) + 1)

            # if self.__debug_on:
            #     print("end-container>>" + ParserHelper.make_value_visible(self.__container_tokens[-1]))
            #     print("split_leading_spaces>>" + ParserHelper.make_value_visible(split_leading_spaces))
            #     print("specific_block_quote_prefix>>:" + ParserHelper.make_value_visible(specific_block_quote_prefix) + ":")
            #     print("fenced-end-error")
            self.report_next_token_error(
                context,
                self.__last_leaf_token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
            )
        self.__last_leaf_token = None
        self.__bq_line_index[num_container_tokens] += 1

    def __handle_link_reference_definition(self, context, token, num_container_tokens):
        scoped_block_quote_token = self.__container_tokens[-1]
        if token.extracted_whitespace:
            column_number_delta = -(
                token.column_number - len(token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("lrd-1-error")
            self.report_next_token_error(
                context, token, column_number_delta=column_number_delta
            )

        found_index = token.link_destination_whitespace.find("\n")
        if found_index != -1 and ParserHelper.is_character_at_index_whitespace(
            token.link_destination_whitespace, found_index + 1
        ):
            line_number_delta = token.link_name_debug.count("\n") + 1

            split_array_index = (
                self.__bq_line_index[num_container_tokens] + line_number_delta
            )
            split_leading_spaces = scoped_block_quote_token.leading_spaces.split("\n")
            specific_block_quote_prefix = split_leading_spaces[split_array_index]

            column_number_delta = -(len(specific_block_quote_prefix) + 1)

            # if self.__debug_on:
            #     print("line_number_delta>>" + str(line_number_delta))
            #     print("split_array_index>>" + str(split_array_index))
            #     print("end-container>>" + ParserHelper.make_value_visible(self.__container_tokens[-1]))
            #     print("split_leading_spaces>>" + ParserHelper.make_value_visible(split_leading_spaces))
            #     print("specific_block_quote_prefix>>:" + ParserHelper.make_value_visible(specific_block_quote_prefix) + ":")
            #     print("lrd-2-error")
            self.report_next_token_error(
                context,
                token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
            )

        found_index = token.link_title_whitespace.find("\n")
        if found_index != -1 and ParserHelper.is_character_at_index_whitespace(
            token.link_title_whitespace, found_index + 1
        ):
            line_number_delta = (
                token.link_name_debug.count("\n")
                + token.link_title_whitespace.count("\n")
                + 1
            )

            split_array_index = (
                self.__bq_line_index[num_container_tokens] + line_number_delta
            )
            split_leading_spaces = scoped_block_quote_token.leading_spaces.split("\n")
            specific_block_quote_prefix = split_leading_spaces[split_array_index]

            column_number_delta = -(len(specific_block_quote_prefix) + 1)
            # if self.__debug_on:
            #     print("line_number_delta>>" + str(line_number_delta))
            #     print("split_array_index>>" + str(split_array_index))
            #     print("end-container>>" + ParserHelper.make_value_visible(self.__container_tokens[-1]))
            #     print("split_leading_spaces>>" + ParserHelper.make_value_visible(split_leading_spaces))
            #     print("specific_block_quote_prefix>>:" + ParserHelper.make_value_visible(specific_block_quote_prefix) + ":")
            #     print("lrd-3-error")
            self.report_next_token_error(
                context,
                token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
            )

        self.__bq_line_index[num_container_tokens] += (
            1
            + token.link_name_debug.count("\n")
            + token.link_destination_whitespace.count("\n")
            + token.link_title_whitespace.count("\n")
            + token.link_title_raw.count("\n")
        )

    def __handle_text(self, context, token, num_container_tokens):
        if self.__last_leaf_token.is_setext_heading:
            scoped_block_quote_token = self.__container_tokens[-1]
            for line_number_delta, next_line in enumerate(
                token.end_whitespace.split("\n")
            ):
                found_index = next_line.find(ParserHelper.whitespace_split_character)
                if found_index != -1:
                    next_line = next_line[0:found_index]
                if next_line:
                    split_leading_spaces = (
                        scoped_block_quote_token.leading_spaces.split("\n")
                    )
                    split_array_index = (
                        self.__bq_line_index[num_container_tokens]
                        + line_number_delta
                        + 1
                    )
                    specific_block_quote_prefix = split_leading_spaces[
                        split_array_index
                    ]
                    calculated_column_number = -(len(specific_block_quote_prefix) + 1)

                    # if self.__debug_on:
                    #     print("split_leading_spaces>>" + ParserHelper.make_value_visible(split_leading_spaces))
                    #     print("split_array_index>>" + ParserHelper.make_value_visible(split_array_index))
                    #     print("specific_block_quote_prefix>>:" + ParserHelper.make_value_visible(specific_block_quote_prefix) + ":")
                    #     print("setext-text-error")
                    self.report_next_token_error(
                        context,
                        token,
                        line_number_delta=line_number_delta,
                        column_number_delta=calculated_column_number,
                    )
            self.__bq_line_index[num_container_tokens] += (
                token.end_whitespace.count("\n") + 1
            )
        elif (
            self.__last_leaf_token.is_html_block or self.__last_leaf_token.is_code_block
        ):
            self.__bq_line_index[num_container_tokens] += (
                token.token_text.count("\n") + 1
            )

    def __handle_paragraph(self, context, token, num_container_tokens):
        self.__last_leaf_token = token
        scoped_block_quote_token = self.__container_tokens[-1]

        for line_number_delta, next_line in enumerate(
            token.extracted_whitespace.split("\n")
        ):
            if next_line:
                # if self.__debug_on:
                #     print("1>" + str(self.__bq_line_index[num_container_tokens]))
                #     print("2>" + str(line_number_delta))
                #     print("3>" + str(scoped_block_quote_token).replace("\n","\\n"))
                split_leading_spaces = scoped_block_quote_token.leading_spaces.split(
                    "\n"
                )
                line_index = (
                    self.__bq_line_index[num_container_tokens] + line_number_delta
                )
                calculated_column_number = len(split_leading_spaces[line_index]) + 1
                # if self.__debug_on:
                #     print("4>" + str(split_leading_spaces).replace("\n","\\n"))
                #     print("5>" + str(line_index))
                #     print("column>" + str(calculated_column_number))
                #     print("para-error")
                self.report_next_token_error(
                    context,
                    scoped_block_quote_token,
                    line_number_delta=line_number_delta,
                    column_number_delta=-calculated_column_number,
                )
        self.__bq_line_index[num_container_tokens] += token.extracted_whitespace.count(
            "\n"
        )

    # pylint: disable=too-many-branches
    def __handle_within_block_quotes(self, context, token):
        num_container_tokens = len(self.__container_tokens)
        # if self.__debug_on:
        #     print(
        #         str(self.__bq_line_index[num_container_tokens])
        #         + "-->token>"
        #         + ParserHelper.make_value_visible(token)
        #     )
        if token.is_paragraph:
            self.__handle_paragraph(context, token, num_container_tokens)
        elif token.is_text:
            self.__handle_text(context, token, num_container_tokens)
        elif token.is_paragraph_end:
            self.__last_leaf_token = None
            self.__bq_line_index[num_container_tokens] += 1
        elif token.is_blank_line:
            self.__handle_blank_line(context, token, num_container_tokens)
        elif token.is_atx_heading:
            self.__handle_atx_heading(context, token, num_container_tokens)
        elif token.is_setext_heading:
            self.__handle_setext_heading(context, token)
        elif token.is_setext_heading_end:
            self.__handle_setext_heading_end(context, token, num_container_tokens)
        elif token.is_html_block or token.is_indented_code_block:
            self.__last_leaf_token = token
        elif (
            token.is_html_block_end
            or token.is_indented_code_block_end
            or token.is_atx_heading_end
        ):
            self.__last_leaf_token = None
        elif token.is_fenced_code_block:
            self.__handle_fenced_code_block(context, token, num_container_tokens)
        elif token.is_fenced_code_block_end:
            self.__handle_fenced_code_block_end(context, token, num_container_tokens)
        elif token.is_thematic_break:
            self.__handle_thematic_break(context, token, num_container_tokens)
        elif token.is_link_reference_definition:
            self.__handle_link_reference_definition(
                context, token, num_container_tokens
            )
        # if self.__debug_on:
        #     print(
        #         str(self.__bq_line_index[num_container_tokens])
        #         + "<--token>"
        #         + ParserHelper.make_value_visible(token)
        #     )

    # pylint: enable=too-many-branches



"""
Module to implement a plugin that ensures that the first line in a file is a top level heading.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd041(Plugin):
    """
    Class to implement a plugin that ensures that the first line in a file is a top level heading.
    """

    def __init__(self):
        super().__init__()
        self.__start_level = None
        self.__front_matter_title = None
        self.__have_seen_first_token = None
        self.__seen_html_block_start = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="first-line-heading,first-line-h1",
            plugin_id="MD041",
            plugin_enabled_by_default=True,
            plugin_description="First line in file should be a top level heading",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md041.md",
            plugin_configuration="level,front_matter_title",
        )

    @classmethod
    def __validate_configuration_level(cls, found_value):
        if found_value < 1 or found_value > 6:
            raise ValueError("Allowable values are between 1 and 6.")

    @classmethod
    def __validate_configuration_title(cls, found_value):
        found_value = found_value.strip()
        if found_value.find(":") != -1:
            raise ValueError("Colons (:) are not allowed in the value.")

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """
        self.__start_level = self.plugin_configuration.get_integer_property(
            "level",
            default_value=1,
            valid_value_fn=self.__validate_configuration_level,
        )
        self.__front_matter_title = (
            self.plugin_configuration.get_string_property(
                "front_matter_title",
                default_value="title",
                valid_value_fn=self.__validate_configuration_title,
            )
            .lower()
            .strip()
        )

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        self.__have_seen_first_token = False
        self.__seen_html_block_start = None

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if not self.__have_seen_first_token:
            if token.is_atx_heading or token.is_setext_heading:
                self.__have_seen_first_token = True
                if token.hash_count != self.__start_level:
                    self.report_next_token_error(context, token)
            elif token.is_front_matter and self.__front_matter_title:
                if self.__front_matter_title in token.matter_map:
                    self.__have_seen_first_token = True
            elif token.is_html_block:
                self.__seen_html_block_start = token
            elif self.__seen_html_block_start:
                assert token.is_text
                html_block_contents = token.token_text.strip()
                if not html_block_contents.startswith(
                    "<h1 "
                ) and not html_block_contents.startswith("<h1>"):
                    self.report_next_token_error(context, self.__seen_html_block_start)
                self.__have_seen_first_token = True
            elif not token.is_blank_line:
                self.report_next_token_error(context, token)
                self.__have_seen_first_token = True



"""
Module to implement a plugin that looks for inconsistent styles for thematic breaks.
"""
from pymarkdown.plugin_manager import Plugin, PluginDetails


class RuleMd035(Plugin):
    """
    Class to implement a plugin that looks for inconsistent styles for thematic breaks.
    """

    __consistent_style = "consistent"

    def __init__(self):
        super().__init__()
        self.__rule_style = None
        self.__actual_style = None

    def get_details(self):
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="hr-style",
            plugin_id="MD035",
            plugin_enabled_by_default=True,
            plugin_description="Horizontal rule style",
            plugin_version="0.5.0",
            plugin_interface_version=1,
            plugin_url="https://github.com/jackdewinter/pymarkdown/blob/main/docs/rules/rule_md035.md",
            plugin_configuration="style",
        )

    @classmethod
    def __validate_configuration_style(cls, found_value):
        if found_value == RuleMd035.__consistent_style:
            return
        if found_value != found_value.strip():
            raise ValueError(
                "Allowable values cannot including leading or trailing spaces."
            )
        is_valid = bool(found_value)
        if is_valid:
            for next_character in found_value:
                if next_character not in " _-*":
                    is_valid = False
                    break
        if not is_valid:
            raise ValueError(
                f"Allowable values are: {RuleMd035.__consistent_style}, '---', '***', or any other horizontal rule text."
            )

    def initialize_from_config(self):
        """
        Event to allow the plugin to load configuration information.
        """
        self.__rule_style = self.plugin_configuration.get_string_property(
            "style",
            default_value=RuleMd035.__consistent_style,
            valid_value_fn=self.__validate_configuration_style,
        )
        if self.__rule_style != RuleMd035.__consistent_style:
            self.__actual_style = self.__rule_style

    def starting_new_file(self):
        """
        Event that the a new file to be scanned is starting.
        """
        if self.__rule_style == RuleMd035.__consistent_style:
            self.__actual_style = None

    def next_token(self, context, token):
        """
        Event that a new token is being processed.
        """
        if token.is_thematic_break:
            if self.__actual_style:
                if self.__actual_style != token.rest_of_line:
                    extra_data = (
                        f"Expected: {self.__actual_style}, Actual: {token.rest_of_line}"
                    )
                    self.report_next_token_error(
                        context, token, extra_error_information=extra_data
                    )
            else:
                self.__actual_style = token.rest_of_line
